# pylint:disable=unused-argument
from typing import TYPE_CHECKING

import claripy
import pyvex
from archinfo.arch_arm import is_arm_arch

from ...errors import SimMemoryMissingError
from ...calling_conventions import SimRegArg, SimStackArg, DefaultCC
from ...engines.vex.claripy.datalayer import value as claripy_value
from ...engines.light import SimEngineLightVEXMixin
from ...knowledge_plugins import Function
from ...storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
from ..typehoon import typevars, typeconsts
from .engine_base import SimEngineVRBase, RichR

if TYPE_CHECKING:
    from .variable_recovery_base import VariableRecoveryStateBase


class SimEngineVRVEX(
    SimEngineLightVEXMixin,
    SimEngineVRBase,
):
    """
    Implements the VEX engine for variable recovery analysis.
    """

    state: "VariableRecoveryStateBase"

    def __init__(self, *args, call_info=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.call_info = call_info or {}

    # Statement handlers

    def _handle_Put(self, stmt):
        offset = stmt.offset
        r = self._expr(stmt.data)
        size = stmt.data.result_size(self.tyenv) // 8

        if offset == self.arch.ip_offset:
            return
        self._assign_to_register(offset, r, size)

    def _handle_Store(self, stmt):
        addr_r = self._expr(stmt.addr)
        size = stmt.data.result_size(self.tyenv) // 8
        r = self._expr(stmt.data)

        self._store(addr_r, r, size, stmt=stmt)

    def _handle_StoreG(self, stmt):
        guard = self._expr(stmt.guard)
        if guard is True:
            addr = self._expr(stmt.addr)
            size = stmt.data.result_size(self.tyenv) // 8
            data = self._expr(stmt.data)
            self._store(addr, data, size, stmt=stmt)

    def _handle_LoadG(self, stmt):
        guard = self._expr(stmt.guard)
        if guard is True:
            addr = self._expr(stmt.addr)
            if addr is not None:
                self.tmps[stmt.dst] = self._load(addr, self.tyenv.sizeof(stmt.dst) // 8)
        elif guard is False:
            data = self._expr(stmt.alt)
            self.tmps[stmt.dst] = data
        else:
            self.tmps[stmt.dst] = None

    def _handle_LLSC(self, stmt: pyvex.IRStmt.LLSC):
        if stmt.storedata is None:
            # load-link
            addr = self._expr(stmt.addr)
            size = self.tyenv.sizeof(stmt.result) // self.arch.byte_width
            data = self._load(addr, size)
            self.tmps[stmt.result] = data
        else:
            # store-conditional
            storedata = self._expr(stmt.storedata)
            addr = self._expr(stmt.addr)
            size = self.tyenv.sizeof(stmt.storedata.tmp) // self.arch.byte_width

            self._store(addr, storedata, size)

            result_size = self.tyenv.sizeof(stmt.result)
            self.tmps[stmt.result] = RichR(claripy.BVV(1, result_size))

    def _handle_NoOp(self, stmt):
        pass

    # Expression handlers

    def _expr(self, expr) -> RichR:
        """

        :param expr:
        :return:
        :rtype: RichR
        """

        r = super()._expr(expr)
        if r is None:
            bits = expr.result_size(self.tyenv)
            return RichR(self.state.top(bits))
        return r

    def _handle_Get(self, expr):
        reg_offset = expr.offset
        reg_size = expr.result_size(self.tyenv) // 8

        # because of how VEX implements MOVCC and MOVCS instructions in ARM THUMB mode, we need to skip the register
        # read if the immediate next instruction is an WrTmp(ITE).
        #
        # MOVCC           R3, #0
        #
        #    46 | ------ IMark(0xfeca2, 2, 1) ------
        #    47 | t299 = CmpLT32U(t8,0x00010000)
        #    48 | t143 = GET:I32(r3)      <-   this read does not exist
        #    49 | t300 = ITE(t299,0x00000000,t143)
        #    50 | PUT(r3) = t300
        #    51 | PUT(pc) = 0x000feca5
        if is_arm_arch(self.arch) and (self.ins_addr & 1) == 1:
            if self.stmt_idx < len(self.block.vex.statements) - 1:
                next_stmt = self.block.vex.statements[self.stmt_idx + 1]
                if isinstance(next_stmt, pyvex.IRStmt.WrTmp) and isinstance(next_stmt.data, pyvex.IRExpr.ITE):
                    return RichR(self.state.top(reg_size * 8))

        return self._read_from_register(reg_offset, reg_size, expr=expr)

    def _handle_Load(self, expr: pyvex.IRExpr.Load) -> RichR:
        addr = self._expr(expr.addr)
        size = expr.result_size(self.tyenv) // 8

        return self._load(addr, size)

    def _handle_CCall(self, expr):  # pylint:disable=useless-return
        # ccalls don't matter
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_Conversion(self, expr: pyvex.IRExpr.Unop) -> RichR:
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    # Function handlers

    def _handle_function_concrete(self, func: Function):
        if func.prototype is None or func.calling_convention is None:
            return

        for arg_loc in func.calling_convention.arg_locs(func.prototype):
            for loc in arg_loc.get_footprint():
                if isinstance(loc, SimRegArg):
                    self._read_from_register(self.arch.registers[loc.reg_name][0] + loc.reg_offset, loc.size)
                elif isinstance(loc, SimStackArg):
                    try:
                        sp: MultiValues = self.state.register_region.load(self.arch.sp_offset, self.arch.bytes)
                    except SimMemoryMissingError:
                        pass
                    else:
                        one_sp = sp.one_value()
                        if one_sp is not None:
                            addr = RichR(loc.stack_offset + one_sp)
                            self._load(addr, loc.size)

    def _process_block_end(self):
        # handles block-end calls
        current_addr = self.state.block_addr
        for target_func in self.call_info.get(current_addr, []):
            self._handle_function_concrete(target_func)

        # handles return statements
        if self.block.vex.jumpkind == "Ijk_Ret":
            # determine the size of the return register
            # TODO: Handle multiple return registers
            cc = self.state.function.calling_convention
            if cc is None:
                cc = DefaultCC[self.arch.name](self.arch)
            if isinstance(cc.RETURN_VAL, SimRegArg):
                ret_val_size = 0
                reg_offset = cc.RETURN_VAL.check_offset(self.arch)
                for i in range(cc.RETURN_VAL.size):
                    try:
                        _ = self.state.register_region.load(reg_offset + i, 1)
                        ret_val_size = i + 1
                    except SimMemoryMissingError:
                        break
                self.state.ret_val_size = (
                    ret_val_size if self.state.ret_val_size is None else max(self.state.ret_val_size, ret_val_size)
                )

    def _handle_Const(self, expr):
        return RichR(
            claripy_value(expr.con.type, expr.con.value, size=expr.con.size), typevar=typeconsts.int_type(expr.con.size)
        )

    def _handle_Add(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data + r1.data, typevar=typeconsts.int_type(result_size), type_constraints=None)

        typevar = None
        if r0.typevar is not None and r1.data.concrete:
            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.AddN(r1.data._model_concrete.value))

        sum_ = r0.data + r1.data
        return RichR(
            sum_,
            typevar=typevar,
            type_constraints={typevars.Subtype(r0.typevar, r1.typevar)},
        )

    def _handle_Sub(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data - r1.data, typevar=typeconsts.int_type(result_size), type_constraints=None)

        typevar = None
        if r0.typevar is not None and r1.data.concrete:
            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.SubN(r1.data._model_concrete.value))

        diff = r0.data - r1.data
        return RichR(
            diff,
            typevar=typevar,
        )

    def _handle_And(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data & r1.data)

        if self.state.is_stack_address(r0.data):
            r = r0.data
        elif self.state.is_stack_address(r1.data):
            r = r1.data
        else:
            r = self.state.top(result_size)
        return RichR(r)

    def _handle_Xor(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data ^ r1.data)

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Or(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data | r1.data)

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Not(self, expr):
        arg = expr.args[0]
        r0 = self._expr(arg)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete:
            # constants
            return RichR(~r0.data)

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Mul(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data * r1.data)

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_DivMod(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            try:
                signed = "U" in expr.op  # Iop_DivModU64to32 vs Iop_DivMod
                from_size = r0.data.size()
                to_size = r1.data.size()
                if signed:
                    quotient = r0.data.SDiv(claripy.SignExt(from_size - to_size, r1.data))
                    remainder = r0.data.SMod(claripy.SignExt(from_size - to_size, r1.data))
                    quotient_size = to_size
                    remainder_size = to_size
                    result = claripy.Concat(
                        claripy.Extract(remainder_size - 1, 0, remainder),
                        claripy.Extract(quotient_size - 1, 0, quotient),
                    )
                else:
                    quotient = r0.data // claripy.ZeroExt(from_size - to_size, r1.data)
                    remainder = r0.data % claripy.ZeroExt(from_size - to_size, r1.data)
                    quotient_size = to_size
                    remainder_size = to_size
                    result = claripy.Concat(
                        claripy.Extract(remainder_size - 1, 0, remainder),
                        claripy.Extract(quotient_size - 1, 0, quotient),
                    )

                return RichR(result)
            except ZeroDivisionError:
                pass

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Div(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            try:
                return RichR(r0.data / r1.data)
            except ZeroDivisionError:
                pass

        r = self.state.top(result_size)
        return RichR(r)

    def _handle_Shr(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(
                claripy.LShR(r0.data, r1.data._model_concrete.value),
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(
            r,
            typevar=r0.typevar,
        )

    def _handle_Sar(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(
                r0.data >> r1.data._model_concrete.value,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(
            r,
            typevar=r0.typevar,
        )

    def _handle_Shl(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(
                r0.data << r1.data._model_concrete.value,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(
            r,
            typevar=r0.typevar,
        )

    def _handle_CmpF(self, expr):
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_16HLto32(self, expr):
        return RichR(self.state.top(32))

    def _handle_Add_v(self, expr, vector_size, vector_count):
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_QSub_v(self, expr, vector_size, vector_count):
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_HAdd_v(self, expr, vector_size, vector_count):
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_Clz(self, expr):
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_Mull(self, expr):
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_CmpEQ(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpNE(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpLE(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpLT(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpGE(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_CmpGT(self, expr):
        arg0, arg1 = expr.args
        _ = self._expr(arg0)
        _ = self._expr(arg1)

        return RichR(self.state.top(1))

    def _handle_Cmp_v(self, expr, vector_size, vector_count):
        return RichR(self.state.top(1))

    def _handle_ExpCmpNE64(self, expr):
        _, _ = self._expr(expr.args[0]), self._expr(expr.args[1])
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_Clz(self, expr):
        arg0 = expr.args[0]
        expr_0 = self._expr(arg0)
        if expr_0 is None:
            return None
        if self.state.is_top(expr_0.data):
            return RichR(self.state.top(expr_0.data.size()))
        return RichR(self.state.top(expr_0.data.size()))

    def _handle_Ctz(self, expr):
        arg0 = expr.args[0]
        expr_0 = self._expr(arg0)
        if expr_0 is None:
            return None
        if self.state.is_top(expr_0.data):
            return RichR(self.state.top(expr_0.data.size()))
        return RichR(self.state.top(expr_0.data.size()))

    _handle_CmpEQ_v = _handle_Cmp_v
    _handle_CmpNE_v = _handle_Cmp_v
    _handle_CmpLE_v = _handle_Cmp_v
    _handle_CmpLT_v = _handle_Cmp_v
    _handle_CmpGE_v = _handle_Cmp_v
    _handle_CmpGT_v = _handle_Cmp_v



import archinfo
import angr
from angr.analyses.cfg import CFGBase


def test_x86_noop_blocks():
    # nop
    arch = archinfo.arch_from_id("x86")
    b = b"\x90\x90\x90\x90\x90\x90\x90\x90"
    p = angr.load_shellcode(b, arch, load_address=0x400000)
    block = p.factory.block(0x400000, opt_level=1, cross_insn_opt=False)
    assert CFGBase._is_noop_block(arch, block) is True
    block = p.factory.block(0x400000, opt_level=1, cross_insn_opt=True)
    assert CFGBase._is_noop_block(arch, block) is True


def test_amd64_noop_blocks():
    # nop
    arch = archinfo.arch_from_id("amd64")
    b = b"\x90\x90\x90\x90\x90\x90\x90\x90"
    p = angr.load_shellcode(b, arch, load_address=0x400000)
    block = p.factory.block(0x400000, opt_level=1, cross_insn_opt=False)
    assert CFGBase._is_noop_block(arch, block) is True
    block = p.factory.block(0x400000, opt_level=1, cross_insn_opt=True)
    assert CFGBase._is_noop_block(arch, block) is True


def test_arm_noop_blocks():
    arch = archinfo.arch_from_id("ARMEL")

    # andeq r0, r0, r0
    b = b"\x00\x00\x00\x00\x00\x00\x00\x00"
    p = angr.load_shellcode(b, arch, load_address=0x400000)
    block = p.factory.block(0x400000, opt_level=1, cross_insn_opt=False)
    assert CFGBase._is_noop_block(arch, block) is True
    block = p.factory.block(0x400000, opt_level=1, cross_insn_opt=True)
    assert CFGBase._is_noop_block(arch, block) is True

    # mov r0, r0
    b = b"\x00\x00\xa0\xe1"
    p = angr.load_shellcode(b, arch, load_address=0x400000)
    block = p.factory.block(0x400000, opt_level=1, cross_insn_opt=False)
    assert CFGBase._is_noop_block(arch, block) is True
    block = p.factory.block(0x400000, opt_level=1, cross_insn_opt=True)
    assert CFGBase._is_noop_block(arch, block) is True


if __name__ == "__main__":
    test_x86_noop_blocks()
    test_amd64_noop_blocks()
    test_arm_noop_blocks()



# pylint:disable=import-outside-toplevel
from typing import Optional, Union

from archinfo import Arch

from .optimization_pass import OptimizationPassStage
from .stack_canary_simplifier import StackCanarySimplifier
from .base_ptr_save_simplifier import BasePointerSaveSimplifier
from .expr_op_swapper import ExprOpSwapper
from .ite_expr_converter import ITEExprConverter
from .lowered_switch_simplifier import LoweredSwitchSimplifier
from .multi_simplifier import MultiSimplifier
from .div_simplifier import DivSimplifier
from .mod_simplifier import ModSimplifier
from .eager_returns import EagerReturnsSimplifier
from .const_derefs import ConstantDereferencesSimplifier
from .register_save_area_simplifier import RegisterSaveAreaSimplifier
from .ret_addr_save_simplifier import RetAddrSaveSimplifier
from .x86_gcc_getpc_simplifier import X86GccGetPcSimplifier


_all_optimization_passes = [
    (RegisterSaveAreaSimplifier, True),
    (StackCanarySimplifier, True),
    (BasePointerSaveSimplifier, True),
    (DivSimplifier, True),
    (MultiSimplifier, True),
    (ModSimplifier, True),
    (ConstantDereferencesSimplifier, True),
    (RetAddrSaveSimplifier, True),
    (X86GccGetPcSimplifier, True),
    (LoweredSwitchSimplifier, False),
    (EagerReturnsSimplifier, True),
    (ITEExprConverter, True),
    (ExprOpSwapper, True),
]


def get_optimization_passes(arch, platform):
    if isinstance(arch, Arch):
        arch = arch.name

    if platform is not None:
        platform = platform.lower()

    passes = []
    for pass_, _ in _all_optimization_passes:
        if (pass_.ARCHES is None or arch in pass_.ARCHES) and (
            pass_.PLATFORMS is None or platform is None or platform in pass_.PLATFORMS
        ):
            passes.append(pass_)

    return passes


def get_default_optimization_passes(arch: Union[Arch, str], platform: Optional[str]):
    if isinstance(arch, Arch):
        arch = arch.name

    if platform is not None:
        platform = platform.lower()

    passes = []
    for pass_, default in _all_optimization_passes:
        if not default:
            continue
        if (pass_.ARCHES is None or arch in pass_.ARCHES) and (
            pass_.PLATFORMS is None or platform is None or platform in pass_.PLATFORMS
        ):
            passes.append(pass_)

    return passes



import angr


def test_unsupported_syscall_simos():
    p = angr.load_shellcode("int 0x80", "x86")
    state = p.factory.entry_state()
    state.regs.eax = 4

    # test that by default trying to perform a syscall without SimUserspace causes the state to go errored
    simgr = p.factory.simulation_manager(state)
    simgr.step()
    assert len(simgr.active) == 1
    simgr.step()
    assert len(simgr.active) == 0
    assert len(simgr.errored) == 1

    # test that when we set BYPASS_UNSUPPORTED_SYSCALLS, we get a syscall stub instead
    state.options.add(angr.options.BYPASS_UNSUPPORTED_SYSCALL)
    simgr = p.factory.simulation_manager(state)
    simgr.step()
    assert len(simgr.active) == 1
    simgr.step()
    assert len(simgr.active) == 1
    assert len(simgr.errored) == 0


if __name__ == "__main__":
    test_unsupported_syscall_simos()



import logging

import pyvex
import claripy

from angr import SimState, load_shellcode
from angr.engines import HeavyVEXMixin
import angr.engines.vex.claripy.ccall as s_ccall

l = logging.getLogger("angr.tests.test_vex")


def test_ccall():
    s = SimState(arch="AMD64")

    l.debug("Testing amd64_actions_ADD")
    l.debug("(8-bit) 1 + 1...")
    arg_l = s.solver.BVV(1, 8)
    arg_r = s.solver.BVV(1, 8)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ADD(s, 8, arg_l, arg_r, 0, platform="AMD64")
    assert s.solver.is_true(cf == 0)
    assert s.solver.is_true(pf == 0)
    assert s.solver.is_true(af == 0)
    assert s.solver.is_true(zf == 0)
    assert s.solver.is_true(sf == 0)
    assert s.solver.is_true(of == 0)

    l.debug("(32-bit) (-1) + (-2)...")
    arg_l = s.solver.BVV(-1, 32)
    arg_r = s.solver.BVV(-1, 32)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ADD(s, 32, arg_l, arg_r, 0, platform="AMD64")
    assert s.solver.is_true(cf == 1)
    assert s.solver.is_true(pf == 0)
    assert s.solver.is_true(af == 1)
    assert s.solver.is_true(zf == 0)
    assert s.solver.is_true(sf == 1)
    assert s.solver.is_true(of == 0)

    l.debug("Testing pc_actions_SUB")
    l.debug(
        "(8-bit) 1 - 1...",
    )
    arg_l = s.solver.BVV(1, 8)
    arg_r = s.solver.BVV(1, 8)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_SUB(s, 8, arg_l, arg_r, 0, platform="AMD64")
    assert s.solver.is_true(cf == 0)
    assert s.solver.is_true(pf == 1)
    assert s.solver.is_true(af == 0)
    assert s.solver.is_true(zf == 1)
    assert s.solver.is_true(sf == 0)
    assert s.solver.is_true(of == 0)

    l.debug("(32-bit) (-1) - (-2)...")
    arg_l = s.solver.BVV(-1, 32)
    arg_r = s.solver.BVV(-2, 32)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_SUB(s, 32, arg_l, arg_r, 0, platform="AMD64")
    assert s.solver.is_true(cf == 0)
    assert s.solver.is_true(pf == 0)
    assert s.solver.is_true(af == 0)
    assert s.solver.is_true(zf == 0)
    assert s.solver.is_true(sf == 0)
    assert s.solver.is_true(of == 0)

    l.debug("Testing pc_actions_ROL")
    l.debug("(8-bit) ROL 1 1...")
    result = s.solver.BVV(2, 8)  # the result of rol(1, 1)
    oldflags = s.solver.BVV(0, 8)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ROL(s, 8, result, None, oldflags, platform="AMD64")
    assert s.solver.is_true(cf == 0)
    assert s.solver.is_true(pf == 0)
    assert s.solver.is_true(af == 0)
    assert s.solver.is_true(zf == 0)
    assert s.solver.is_true(sf == 0)
    assert s.solver.is_true(of == 0)

    l.debug("(32-bit) ROL (-1) (-2)... (shift out of range)")
    result = s.solver.BVV(-1, 32)  # the result of rol(-1, 0xfe)
    oldflags = s.solver.BVV(0, 32)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ROL(s, 32, result, None, oldflags, platform="AMD64")
    assert s.solver.is_true(cf == 1)
    assert s.solver.is_true(pf == 0)
    assert s.solver.is_true(af == 0)
    assert s.solver.is_true(zf == 0)
    assert s.solver.is_true(sf == 0)
    assert s.solver.is_true(of == 0)

    l.debug("Testing pc_actions_ROR")
    l.debug("(32-bit) ROR 2 1...")
    result = s.solver.BVV(1, 32)  # the result of ror(2, 1)
    oldflags = s.solver.BVV(0, 8)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ROR(s, 32, result, None, oldflags, platform="AMD64")
    assert s.solver.is_true(cf == 0)
    assert s.solver.is_true(pf == 0)
    assert s.solver.is_true(af == 0)
    assert s.solver.is_true(zf == 0)
    assert s.solver.is_true(sf == 0)
    assert s.solver.is_true(of == 0)

    l.debug("Testing pc_actions_ROR")
    l.debug("(32-bit) ROR 1 1...")
    result = s.solver.BVV(0x80000000, 32)  # the result of ror(1, 1)
    oldflags = s.solver.BVV(0, 8)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ROR(s, 32, result, None, oldflags, platform="AMD64")
    assert s.solver.is_true(cf == 1)
    assert s.solver.is_true(pf == 0)
    assert s.solver.is_true(af == 0)
    assert s.solver.is_true(zf == 0)
    assert s.solver.is_true(sf == 0)
    assert s.solver.is_true(of == 1)

    l.debug("Testing pc_actions_ROR")
    l.debug("(32-bit) ROR -1 1...")
    result = s.solver.BVV(-1, 32)  # the result of ror(0xffffffff, 1)
    oldflags = s.solver.BVV(0, 32)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ROR(s, 32, result, None, oldflags, platform="AMD64")
    assert s.solver.is_true(cf == 1)
    assert s.solver.is_true(pf == 0)
    assert s.solver.is_true(af == 0)
    assert s.solver.is_true(zf == 0)
    assert s.solver.is_true(sf == 0)
    assert s.solver.is_true(of == 0)

    l.debug("(32-bit) ROR (-1) (-2)... (shift out of range)")
    result = s.solver.BVV(-1, 32)  # the result of ror(0xffffffff, 0xfe)
    oldflags = s.solver.BVV(0, 32)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ROR(s, 32, result, None, oldflags, platform="AMD64")
    assert s.solver.is_true(cf == 1)
    assert s.solver.is_true(pf == 0)
    assert s.solver.is_true(af == 0)
    assert s.solver.is_true(zf == 0)
    assert s.solver.is_true(sf == 0)
    assert s.solver.is_true(of == 0)

    l.debug("Testing amd64_actions_ADCX")

    l.debug("(ADCX, 32-bit) 0xffffffff + 1...")
    arg_l = s.solver.BVV(0xFFFFFFFF, 32)
    arg_r = s.solver.BVV(1, 32)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ADCX(s, 32, arg_l, arg_r, s.solver.BVV(0, 32), True, platform="AMD64")
    assert s.solver.is_true(cf == 1)
    assert s.solver.is_true(of == 0)

    l.debug("(ADOX, 32-bit) 0xffffffff + 1...")
    arg_l = s.solver.BVV(0xFFFFFFFF, 32)
    arg_r = s.solver.BVV(1, 32)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ADCX(s, 32, arg_l, arg_r, s.solver.BVV(0, 32), False, platform="AMD64")
    assert s.solver.is_true(cf == 0)
    assert s.solver.is_true(of == 1)

    l.debug("(ADCX, 64-bit) 0xffffffffffffffff + 1...")
    arg_l = s.solver.BVV(0xFFFFFFFFFFFFFFFF, 64)
    arg_r = s.solver.BVV(1, 64)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ADCX(s, 64, arg_l, arg_r, s.solver.BVV(0, 64), True, platform="AMD64")
    assert s.solver.is_true(cf == 1)
    assert s.solver.is_true(of == 0)

    l.debug("(ADOX, 64-bit) 0xffffffffffffffff + 1...")
    arg_l = s.solver.BVV(0xFFFFFFFFFFFFFFFF, 64)
    arg_r = s.solver.BVV(1, 64)
    cf, pf, af, zf, sf, of = s_ccall.pc_actions_ADCX(s, 64, arg_l, arg_r, s.solver.BVV(0, 64), False, platform="AMD64")
    assert s.solver.is_true(cf == 0)
    assert s.solver.is_true(of == 1)


def test_aarch64_32bit_ccalls():
    # GitHub issue #1238
    s = SimState(arch="AArch64")

    x = s.solver.BVS("x", 32)
    # A normal operation
    flag_z = s_ccall.arm64g_calculate_flag_z(s, s_ccall.ARM64G_CC_OP_ADD32, x, s.solver.BVV(1, 32), 0)
    assert s.satisfiable(extra_constraints=(flag_z == 0,))
    assert s.satisfiable(extra_constraints=(flag_z == 1,))
    # What VEX does
    flag_z = s_ccall.arm64g_calculate_flag_z(s, s_ccall.ARM64G_CC_OP_ADD32, x.zero_extend(32), s.solver.BVV(1, 64), 0)
    assert s.satisfiable(extra_constraints=(flag_z == 0,))
    assert s.satisfiable(extra_constraints=(flag_z == 1,))


def test_some_vector_ops():
    engine = HeavyVEXMixin(None)
    s = SimState(arch="AMD64")

    def translate(state, op, args):
        return engine._perform_vex_expr_Op(op, args)

    a = s.solver.BVV(0xFFFF0000000100020003000400050006, 128)
    b = s.solver.BVV(0x00020002000200020002000200020002, 128)

    calc_result = translate(s, "Iop_Sub16x8", (a, b))
    correct_result = s.solver.BVV(0xFFFDFFFEFFFF00000001000200030004, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_CmpEQ16x8", (a, b))
    correct_result = s.solver.BVV(0x000000000000FFFF0000000000000000, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_CmpEQ8x16", (a, b))
    correct_result = s.solver.BVV(0x0000FF00FF00FFFFFF00FF00FF00FF00, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_CmpGT16Sx8", (a, b))
    correct_result = s.solver.BVV(0x0000000000000000FFFFFFFFFFFFFFFF, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_CmpGT16Ux8", (a, b))
    correct_result = s.solver.BVV(0xFFFF000000000000FFFFFFFFFFFFFFFF, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_InterleaveLO16x8", (a, b))
    correct_result = s.solver.BVV(0x00030002000400020005000200060002, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_InterleaveLO8x16", (a, b))
    correct_result = s.solver.BVV(0x00000302000004020000050200000602, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_Min8Ux16", (a, b))
    correct_result = s.solver.BVV(0x00020000000100020002000200020002, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_Min8Sx16", (a, b))
    correct_result = s.solver.BVV(0xFFFF0000000100020002000200020002, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_QNarrowBin16Sto8Ux16", (a, b))
    correct_result = s.solver.BVV(0x00000102030405060202020202020202, 128)
    assert s.solver.is_true(calc_result == correct_result)

    c = s.solver.BVV(0xFF008877, 32)
    d = s.solver.BVV(0x11111111, 32)

    calc_result = translate(s, "Iop_HAdd8Sx4", (c, d))
    correct_result = s.solver.BVV(0x0808CC44, 32)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_QAdd8Sx4", (c, d))
    correct_result = s.solver.BVV(0x1011997F, 32)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_QAdd8Ux4", (c, d))
    correct_result = s.solver.BVV(0xFF119988, 32)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_QSub8Sx4", (c, d))
    correct_result = s.solver.BVV(0xEEEF8066, 32)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_QSub8Ux4", (c, d))
    correct_result = s.solver.BVV(0xEE007766, 32)
    assert s.solver.is_true(calc_result == correct_result)

    e = s.solver.BVV(0xFF00887766554433, 64)
    f = s.solver.BVV(0x0202000200020002, 64)

    calc_result = translate(s, "Iop_QNarrowBin16Sto8Ux8", (e, f))
    correct_result = s.solver.BVV(0x0000FFFFFF020202, 64)
    assert s.solver.is_true(calc_result == correct_result)

    gg = claripy.BVV(0x111111112222222233333333FFFFFFFF, 128)
    h = claripy.BVV(0x1111111100000000FFFFFFFFFFFFFFFF, 128)

    calc_result = translate(s, "Iop_CmpEQ32Fx4", (gg, h))
    correct_result = claripy.BVV(0xFFFFFFFF000000000000000000000000, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_Clz32x4", (gg,))
    correct_result = claripy.BVV(0x00000003000000020000000200000000, 128)
    assert s.solver.is_true(calc_result == correct_result)

    i = claripy.BVV(0x1001000000001000, 64)
    j = claripy.BVV(0x100000000000F000, 64)

    calc_result = translate(s, "Iop_Mull16Sx4", (i, j))
    correct_result = claripy.BVV(0x10010000000000000000000FF000000, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_Mull16Ux4", (i, j))
    correct_result = claripy.BVV(0x100100000000000000000000F000000, 128)
    assert s.solver.is_true(calc_result == correct_result)

    k = claripy.BVV(0xE7, 8)
    ll = claripy.BVV(0x1234, 16)
    m = claripy.BVV(0x12345678, 32)

    calc_result = translate(s, "Iop_Dup8x8", (k,))
    correct_result = claripy.BVV(0xE7E7E7E7E7E7E7E7, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_Dup8x16", (k,))
    correct_result = claripy.BVV(0xE7E7E7E7E7E7E7E7E7E7E7E7E7E7E7E7, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_Dup16x4", (ll,))
    correct_result = claripy.BVV(0x1234123412341234, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_Dup16x8", (ll,))
    correct_result = claripy.BVV(0x12341234123412341234123412341234, 128)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_Dup32x2", (m,))
    correct_result = claripy.BVV(0x1234567812345678, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_Dup32x4", (m,))
    correct_result = claripy.BVV(0x12345678123456781234567812345678, 128)
    assert s.solver.is_true(calc_result == correct_result)

    n = claripy.BVV(0x0123456789ABCDEF, 64)
    o = claripy.BVV(0xAF, 8)
    p = claripy.BVV(0xDFEC, 16)
    q = claripy.BVV(0xBFCFDFEF, 32)
    r = claripy.BVV(0x0102030405060708, 64)
    ss = claripy.BVS("index", 8)

    # According to the source code of LibVex, the index is a U8 constant
    calc_result = translate(s, "Iop_GetElem8x8", (n, claripy.BVV(0, 8)))
    correct_result = claripy.BVV(0xEF, 8)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_GetElem8x8", (n, claripy.BVV(1, 8)))
    correct_result = claripy.BVV(0xCD, 8)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_GetElem8x8", (n, claripy.BVV(6, 8)))
    correct_result = claripy.BVV(0x23, 8)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_GetElem8x8", (n, claripy.BVV(7, 8)))
    correct_result = claripy.BVV(0x01, 8)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_GetElem16x4", (n, claripy.BVV(0, 8)))
    correct_result = claripy.BVV(0xCDEF, 16)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_GetElem16x4", (n, claripy.BVV(3, 8)))
    correct_result = claripy.BVV(0x0123, 16)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_GetElem32x2", (n, claripy.BVV(0, 8)))
    correct_result = claripy.BVV(0x89ABCDEF, 32)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_GetElem32x2", (n, claripy.BVV(1, 8)))
    correct_result = claripy.BVV(0x01234567, 32)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_SetElem8x8", (n, claripy.BVV(0, 8), o))
    correct_result = claripy.BVV(0x0123456789ABCDAF, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_SetElem8x8", (n, claripy.BVV(1, 8), o))
    correct_result = claripy.BVV(0x0123456789ABAFEF, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_SetElem8x8", (n, claripy.BVV(6, 8), o))
    correct_result = claripy.BVV(0x01AF456789ABCDEF, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_SetElem8x8", (n, claripy.BVV(7, 8), o))
    correct_result = claripy.BVV(0xAF23456789ABCDEF, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_SetElem16x4", (n, claripy.BVV(0, 8), p))
    correct_result = claripy.BVV(0x0123456789ABDFEC, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_SetElem16x4", (n, claripy.BVV(3, 8), p))
    correct_result = claripy.BVV(0xDFEC456789ABCDEF, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_SetElem32x2", (n, claripy.BVV(0, 8), q))
    correct_result = claripy.BVV(0x01234567BFCFDFEF, 64)
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_SetElem32x2", (n, claripy.BVV(1, 8), q))
    correct_result = claripy.BVV(0xBFCFDFEF89ABCDEF, 64)
    assert s.solver.is_true(calc_result == correct_result)

    # Symbolic indexes
    calc_result = translate(ss, "Iop_GetElem8x8", (r, ss))
    correct_result = claripy.If(
        ss == 7,
        claripy.BVV(0x01, 8),
        claripy.If(
            ss == 6,
            claripy.BVV(0x02, 8),
            claripy.If(
                ss == 5,
                claripy.BVV(0x03, 8),
                claripy.If(
                    ss == 4,
                    claripy.BVV(0x04, 8),
                    claripy.If(
                        ss == 3,
                        claripy.BVV(0x05, 8),
                        claripy.If(
                            ss == 2,
                            claripy.BVV(0x06, 8),
                            claripy.If(ss == 1, claripy.BVV(0x07, 8), claripy.BVV(0x08, 8)),
                        ),
                    ),
                ),
            ),
        ),
    )
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_GetElem16x4", (r, ss))
    correct_result = claripy.If(
        ss == 3,
        claripy.BVV(0x0102, 16),
        claripy.If(
            ss == 2,
            claripy.BVV(0x0304, 16),
            claripy.If(ss == 1, claripy.BVV(0x0506, 16), claripy.BVV(0x0708, 16)),
        ),
    )
    assert s.solver.is_true(calc_result == correct_result)

    calc_result = translate(s, "Iop_GetElem32x2", (r, ss))
    correct_result = claripy.If(ss == 1, claripy.BVV(0x01020304, 32), claripy.BVV(0x05060708, 32))
    assert s.solver.is_true(calc_result == correct_result)

    r_slices = r.chop(8)
    calc_result = translate(s, "Iop_SetElem8x8", (r, ss, o))
    correct_result = claripy.Concat(
        claripy.If(ss == 7, o, r_slices[0]),
        claripy.If(ss == 6, o, r_slices[1]),
        claripy.If(ss == 5, o, r_slices[2]),
        claripy.If(ss == 4, o, r_slices[3]),
        claripy.If(ss == 3, o, r_slices[4]),
        claripy.If(ss == 2, o, r_slices[5]),
        claripy.If(ss == 1, o, r_slices[6]),
        claripy.If(ss == 0, o, r_slices[7]),
    )
    assert s.solver.is_true(calc_result == correct_result)

    r_slices = r.chop(16)
    calc_result = translate(s, "Iop_SetElem16x4", (r, ss, p))
    correct_result = claripy.Concat(
        claripy.If(ss == 3, p, r_slices[0]),
        claripy.If(ss == 2, p, r_slices[1]),
        claripy.If(ss == 1, p, r_slices[2]),
        claripy.If(ss == 0, p, r_slices[3]),
    )
    assert s.solver.is_true(calc_result == correct_result)

    r_slices = r.chop(32)
    calc_result = translate(s, "Iop_SetElem32x2", (r, ss, q))
    correct_result = claripy.Concat(
        claripy.If(ss == 1, q, r_slices[0]),
        claripy.If(ss == 0, q, r_slices[1]),
    )
    assert s.solver.is_true(calc_result == correct_result)


def test_store_simplification():
    state = SimState(arch="X86")
    state.regs.esp = state.solver.BVS("stack_pointer", 32)
    state.regs.ebp = state.solver.BVS("base_pointer", 32)
    state.regs.eax = state.solver.BVS("base_eax", 32)

    irsb = pyvex.IRSB(b"PT]\xc2\x10\x00", 0x4000, state.arch)
    sim_successors = HeavyVEXMixin(None).process(state.copy(), irsb=irsb)
    exit_state = sim_successors.all_successors[0]

    assert claripy.backends.z3.is_true(exit_state.regs.ebp == state.regs.esp - 4)


def test_loadg_no_constraint_creation():
    state = SimState(arch="armel", mode="symbolic")
    engine = HeavyVEXMixin(None)

    stmt = pyvex.IRStmt.LoadG(
        "Iend_LE",
        "ILGop_16Uto32",
        0,  # dst
        pyvex.IRExpr.Const(pyvex.const.U32(0x2000)),  # addr (src)
        pyvex.IRExpr.Const(pyvex.const.U32(0x1337)),  # alt
        pyvex.IRExpr.RdTmp(1),  # guard
    )
    tyenv = pyvex.IRTypeEnv(state.arch)
    tyenv.types = ["Ity_I32", "Ity_I32"]
    state.scratch.set_tyenv(tyenv)
    state.scratch.temps[1] = state.solver.BVS("tmp_1", 32)
    engine.state = state
    engine._handle_vex_stmt(stmt)

    # LOADG should not create new constraints - it is a simple conditional memory read. The conditions should only be
    # used inside the value AST to guard the memory read.
    assert not state.solver.constraints
    assert state.scratch.temps[0] is not None
    assert state.scratch.temps[0].variables.issuperset(state.scratch.temps[1].variables)
    assert state.scratch.temps[0].op == "If"


def test_amd64_ud012_behaviors():
    # Test if VEX's lifter behaves as what CFGFast expects
    #
    # Note: if such behaviors change in the future, you also need to fix the ud{0,1,2} handling logic in
    # CFGFast._generate_cfgnode().

    # according to VEX, ud0 is not part of the block
    a = load_shellcode(b"\x90\x90\x0f\xff", "amd64")
    block_0 = a.factory.block(0)
    assert block_0.size == 2

    # according to VEX, ud1 is not part of the block
    a = load_shellcode(b"\x90\x90\x0f\xb9", "amd64")
    block_1 = a.factory.block(0)
    assert block_1.size == 2

    # according to VEX, ud2 under AMD64 *is* part of the block
    a = load_shellcode(b"\x90\x90\x0f\x0b", "amd64")
    block_2 = a.factory.block(0)
    assert block_2.size == 4


def test_x86_ud2_behaviors():
    # Test if VEX's lifter behaves as what CFGFast expects
    #
    # Note: if such behaviors change in the future, you also need to fix the ud2 handling logic in
    # CFGFast._generate_cfgnode().

    # according to VEX, ud2 on x86 is not part of the block
    a = load_shellcode(b"\x90\x90\x0f\x0b", "x86")
    block_0 = a.factory.block(0)
    assert block_0.size == 2


def test_blsr():
    p = load_shellcode(bytes.fromhex("c4e2f8f3cf0f95c00fb6c0c3"), arch="amd64")
    # compiled form of !!(x & (x - 1))
    # 2c0:   c4 e2 f8 f3 cf          blsr   %rdi,%rax

    # 2c5:   0f 95 c0                setne  %al
    # 2c8:   0f b6 c0                movzbl %al,%eax
    # 2cb:   c3                      retq
    x = claripy.BVS("x", 64)
    s = p.factory.call_state(0, x)
    sm = p.factory.simulation_manager(s)
    sm.run()

    target_func = claripy.If((x & (x - 1)) == 0, claripy.BVV(0, 64), 1)
    solver = claripy.Solver()
    solver.add(sm.one_deadended.regs.rax != target_func)
    assert not solver.satisfiable()


if __name__ == "__main__":
    g = globals().copy()
    for func_name, func in g.items():
        if func_name.startswith("test_") and hasattr(func, "__call__"):
            func()



# pylint: disable=missing-class-docstring
from ailment.expression import BinaryOp, Convert, Const

from .base import PeepholeOptimizationExprBase


class RemoveRedundantConversions(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "Remove redundant conversions around binary operators"
    expr_classes = (BinaryOp,)

    def optimize(self, expr: BinaryOp):
        # TODO make this lhs/rhs agnostic
        if isinstance(expr.operands[0], Convert):
            # check: is the lhs convert an up-cast and is rhs a const?
            if expr.operands[0].to_bits > expr.operands[0].from_bits and isinstance(expr.operands[1], Const):
                to_bits = expr.operands[0].to_bits
                from_bits = expr.operands[0].from_bits
                if expr.op == "And":
                    if 0 <= expr.operands[1].value <= ((1 << from_bits) - 1) or expr.operands[1].value >= (
                        1 << to_bits
                    ) - (1 << (from_bits - 1)):
                        con = Const(None, None, expr.operands[1].value, from_bits, **expr.operands[1].tags)
                        new_expr = BinaryOp(
                            expr.idx, "And", (expr.operands[0].operand, con), expr.signed, bits=from_bits, **expr.tags
                        )
                        return Convert(
                            expr.operands[0].idx,
                            from_bits,
                            to_bits,
                            expr.operands[0].is_signed,
                            new_expr,
                            **expr.operands[0].tags,
                        )

                elif expr.op in {
                    "CmpEQ",
                    "CmpNE",
                    "CmpGT",
                    "CmpGE",
                    "CmpGTs",
                    "CmpGEs",
                    "CmpLT",
                    "CmpLE",
                    "CmpLTs",
                    "CmpLEs",
                }:
                    if 0 <= expr.operands[1].value <= ((1 << from_bits) - 1) or (
                        expr.operands[0].is_signed and expr.operands[1].value >= (1 << to_bits) - (1 << (from_bits - 1))
                    ):
                        con = Const(None, None, expr.operands[1].value, from_bits, **expr.operands[1].tags)
                        new_expr = BinaryOp(
                            expr.idx, expr.op, (expr.operands[0].operand, con), expr.signed, bits=from_bits, **expr.tags
                        )
                        return new_expr

            elif (
                isinstance(expr.operands[1], Convert)
                and expr.operands[1].to_bits == expr.operands[0].to_bits
                and expr.operands[1].from_bits == expr.operands[0].from_bits
            ):
                if expr.op in {"Add", "Sub"}:
                    op0 = expr.operands[0]
                    op0_inner = expr.operands[0].operand
                    # op1 = expr.operands[1]
                    op1_inner = expr.operands[1].operand

                    new_expr = BinaryOp(
                        expr.idx,
                        expr.op,
                        (op0_inner, op1_inner),
                        expr.signed,
                        bits=op0.from_bits,
                        **expr.tags,
                    )
                    r = Convert(
                        expr.idx,
                        op0.from_bits,
                        op0.to_bits,
                        op0.is_signed,
                        new_expr,
                        **op0.tags,
                    )
                    return r

        return None



import angr
import glob
import os

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_stack_end():
    for fn in glob.glob(os.path.join(test_location, "*", "fauxware")):
        p = angr.Project(fn, auto_load_libs=False)

        # normal state
        s = p.factory.full_init_state()
        offset = s.solver.eval(p.arch.initial_sp - s.regs.sp)

        # different stack ends
        for n in [0x1337000, 0xBAAAAA00, 0x100, 0xFFFFFF00, 0x13371337000, 0xBAAAAAAA0000, 0xFFFFFFFFFFFFFF00]:
            if n.bit_length() > p.arch.bits:
                continue
            s = p.factory.full_init_state(stack_end=n)
            assert s.solver.eval_one(s.regs.sp + offset == n)


def test_execstack():
    bin_path = os.path.join(test_location, "x86_64", "fauxware")
    proj = angr.Project(bin_path, auto_load_libs=False)
    # manually mark the stack as executable
    proj.loader.main_object.execstack = True
    s = proj.factory.blank_state()
    assert s.memory._stack_perms == 7


def test_brk():
    for fn in glob.glob(os.path.join(test_location, "*", "fauxware")):
        p = angr.Project(fn, auto_load_libs=False)

        # different stack ends
        for n in [0x1337000, 0xBAAAAA00, 0x100, 0xFFFFFF00, 0x13371337000, 0xBAAAAAAA0000, 0xFFFFFFFFFFFFFF00]:
            if n.bit_length() > p.arch.bits:
                continue
            s = p.factory.full_init_state(brk=n)
            assert s.solver.eval_one(s.posix.brk == n)


if __name__ == "__main__":
    test_stack_end()
    test_execstack()
    test_brk()



import os
import unittest

import angr
import angr.analyses.decompiler


# pylint: disable=missing-class-docstring
# pylint: disable=no-self-use
class TestClinic(unittest.TestCase):
    def test_smoketest(self):
        binary_path = os.path.join(
            os.path.dirname(os.path.realpath(str(__file__))), "..", "..", "binaries", "tests", "x86_64", "all"
        )
        proj = angr.Project(binary_path, auto_load_libs=False, load_debug_info=True)

        cfg = proj.analyses.CFG(normalize=True)
        main_func = cfg.kb.functions["main"]

        proj.analyses.Clinic(main_func)


if __name__ == "__main__":
    unittest.main()



import os


import angr
import claripy as cp

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def regression_test_memcmp_strlen_simprocedure_interaction():
    # import logging
    # logging.getLogger('angr.manager').setLevel(logging.DEBUG)

    bin_path = os.path.join(test_location, "i386", "cpp_regression_test_ch25")
    p = angr.Project(bin_path, auto_load_libs=True)  # this binary requires the loading of libstdc++.so.6
    argv1 = cp.Concat(*[cp.BVS("argv%d" % i, 8) for i in range(48)])

    state = p.factory.full_init_state(args=[bin_path, argv1], add_options=angr.sim_options.unicorn)

    sm = p.factory.simulation_manager(state)
    x = sm.explore(find=0x8048B9B, num_find=3)

    assert len(x.found) == 1
    for state in x.found:
        solution = state.solver.eval_one(argv1, cast_to=bytes).strip(b"\x00")
        assert solution == b"Here_you_have_to_understand_a_little_C++_stuffs"


if __name__ == "__main__":
    regression_test_memcmp_strlen_simprocedure_interaction()



import angr

######################################
# listen (but not really)
######################################
import logging

l = logging.getLogger(name=__name__)


class listen(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, sockfd, backlog):  # pylint:disable=unused-argument
        return self.state.solver.Unconstrained("listen", self.arch.sizeof["int"], key=("api", "listen"))



import os
import angr
from angr.codenode import BlockNode, HookNode, SyscallNode

BIN_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries")


def test_ret_float():
    class F1(angr.SimProcedure):
        def run(self):
            return 12.5

    p = angr.load_shellcode(b"X", arch="i386")

    p.hook(0x1000, F1(prototype="float (x)();"))
    p.hook(0x2000, F1(prototype="double (x)();"))

    s = p.factory.call_state(addr=0x1000, ret_addr=0, prototype="float(x)()")
    succ = s.step()
    assert len(succ.successors) == 1
    s2 = succ.flat_successors[0]
    assert not s2.regs.st0.symbolic
    assert s2.solver.eval(s2.regs.st0.raw_to_fp()) == 12.5

    s = p.factory.call_state(addr=0x2000, ret_addr=0, prototype="double(x)()")
    succ = s.step()
    assert len(succ.successors) == 1
    s2 = succ.flat_successors[0]
    assert not s2.regs.st0.symbolic
    assert s2.solver.eval(s2.regs.st0.raw_to_fp()) == 12.5

    p = angr.load_shellcode(b"X", arch="amd64")

    p.hook(0x1000, F1(prototype="float (x)();"))
    p.hook(0x2000, F1(prototype="double (x)();"))

    s = p.factory.call_state(addr=0x1000, ret_addr=0, prototype="float(x)()")
    succ = s.step()
    assert len(succ.successors) == 1
    s2 = succ.flat_successors[0]
    res = s2.registers.load("xmm0", 4).raw_to_fp()
    assert not res.symbolic
    assert s2.solver.eval(res) == 12.5

    s = p.factory.call_state(addr=0x2000, ret_addr=0, prototype="double(x)()")
    succ = s.step()
    assert len(succ.successors) == 1
    s2 = succ.flat_successors[0]
    res = s2.registers.load("xmm0", 8).raw_to_fp()
    assert not res.symbolic
    assert s2.solver.eval(res) == 12.5


def test_syscall_and_simprocedure():
    bin_path = os.path.join(BIN_PATH, "tests", "cgc", "CADET_00002")
    proj = angr.Project(bin_path, auto_load_libs=False)
    cfg = proj.analyses.CFGFast(normalize=True)

    # check syscall
    node = cfg.get_any_node(proj.loader.kernel_object.mapped_base + 1)
    func = proj.kb.functions[node.addr]

    assert node.is_simprocedure
    assert node.is_syscall
    assert not node.to_codenode().is_hook
    assert not proj.is_hooked(node.addr)
    assert func.is_syscall
    assert func.is_simprocedure
    assert type(proj.factory.snippet(node.addr)) == SyscallNode

    # check normal functions
    node = cfg.get_any_node(0x80480A0)
    func = proj.kb.functions[node.addr]

    assert not node.is_simprocedure
    assert not node.is_syscall
    assert not proj.is_hooked(node.addr)
    assert not func.is_syscall
    assert not func.is_simprocedure
    assert type(proj.factory.snippet(node.addr)) == BlockNode

    # check hooked functions
    proj.hook(0x80480A0, angr.SIM_PROCEDURES["libc"]["puts"]())
    cfg = proj.analyses.CFGFast(normalize=True)  # rebuild cfg to updated nodes
    node = cfg.get_any_node(0x80480A0)
    func = proj.kb.functions[node.addr]

    assert node.is_simprocedure
    assert not node.is_syscall
    assert proj.is_hooked(node.addr)
    assert not func.is_syscall
    assert func.is_simprocedure
    assert type(proj.factory.snippet(node.addr)) == HookNode


def test_inet_ntoa() -> None:
    """
    Test the inet_ntoa simprocedure for functionality
    """
    bin_path = os.path.join(BIN_PATH, "tests", "x86_64", "inet_ntoa")
    proj = angr.Project(bin_path, auto_load_libs=False, use_sim_procedures=True)
    initial_state = proj.factory.entry_state()
    simgr = proj.factory.simgr(initial_state)
    after = simgr.run()
    assert after.deadended[0].posix.dumps(1) == b"192.168.192.168\n"


if __name__ == "__main__":
    test_ret_float()
    test_syscall_and_simprocedure()
    test_inet_ntoa()



from typing import Tuple, Dict, Set, DefaultDict, Any, Optional
from collections import defaultdict

from angr.serializable import Serializable


class PropagationModel(Serializable):
    """
    This class stores the propagation result that comes out of Propagator.
    """

    __slots__ = (
        "key",
        "node_iterations",
        "states",
        "block_initial_reg_values",
        "replacements",
        "equivalence",
        # internals of the function graph visitor
        "graph_visitor",
    )

    def __init__(
        self,
        prop_key: Tuple,
        node_iterations: Optional[DefaultDict[Any, int]] = None,
        states: Optional[Dict] = None,
        block_initial_reg_values: Optional[Dict] = None,
        replacements: Optional[DefaultDict[Any, Dict]] = None,
        equivalence: Optional[Set] = None,
    ):
        self.key = prop_key
        self.node_iterations = node_iterations if node_iterations is not None else defaultdict(int)
        self.states = states if states is not None else {}
        self.block_initial_reg_values = block_initial_reg_values if block_initial_reg_values is not None else {}
        self.replacements = replacements
        self.equivalence = equivalence if equivalence is not None else set()

        self.graph_visitor = None

    def downsize(self):
        self.node_iterations = None
        self.block_initial_reg_values = None
        self.states = None
        self.graph_visitor = None



import angr

import logging

l = logging.getLogger("angr_tests")

import os

test_location = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries", "tests")


def test_strncpy():
    # auto_load_libs can't be disabled as the test cases failed.
    strncpy_amd64 = angr.Project(
        os.path.join(test_location, "x86_64", "strncpy"),
        load_options={"auto_load_libs": True},
        exclude_sim_procedures_list=["strncpy"],
    )
    explorer = strncpy_amd64.factory.simulation_manager()
    explorer.explore(find=[0x4005FF])
    s = explorer.found[0]
    result = s.solver.eval(s.memory.load(s.regs.rax, 16), cast_to=bytes)
    assert result == b"why hello there\x00"


def test_strncpy_size():
    # auto_load_libs can't be disabled as the test cases failed.
    strncpy_size_amd64 = angr.Project(
        os.path.join(test_location, "x86_64", "strncpy-size"),
        load_options={"auto_load_libs": True},
        exclude_sim_procedures_list=["strncpy"],
    )
    explorer = strncpy_size_amd64.factory.simulation_manager()
    cfg = strncpy_size_amd64.analyses.CFG(objects=[strncpy_size_amd64.loader.main_object], normalize=True)
    explorer.use_technique(angr.exploration_techniques.LoopSeer(cfg=cfg, bound=50))
    explorer.explore(find=[0x40064C])
    s = explorer.found[0]
    result = s.solver.eval(s.memory.load(s.regs.rax, 40), cast_to=bytes)
    assert (
        result
        == b"just testing things\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    )


def test_strncpy_verify_null():
    # auto_load_libs can't be disabled as the test cases failed.
    strncpy_verify_null_amd64 = angr.Project(
        os.path.join(test_location, "x86_64", "strncpy-verify-null"),
        load_options={"auto_load_libs": True},
        exclude_sim_procedures_list=["strncpy"],
    )
    explorer = strncpy_verify_null_amd64.factory.simulation_manager()
    cfg = strncpy_verify_null_amd64.analyses.CFG(objects=[strncpy_verify_null_amd64.loader.main_object], normalize=True)
    explorer.use_technique(angr.exploration_techniques.LoopSeer(cfg=cfg, bound=50))
    explorer.explore(find=[0x40064C])
    s = explorer.found[0]
    result = s.solver.eval(s.memory.load(s.regs.rax, 40), cast_to=bytes)
    assert result == b"just testing things\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00AAAAAA\x00"


def test_strstr_and_strncpy():
    # auto_load_libs can't be disabled as the test cases failed.
    strstr_and_strncpy_amd64 = angr.Project(
        os.path.join(test_location, "x86_64", "strstr_and_strncpy"),
        load_options={"auto_load_libs": True},
        exclude_sim_procedures_list=["strstr"],
    )
    explorer = strstr_and_strncpy_amd64.factory.simulation_manager()
    cfg = strstr_and_strncpy_amd64.analyses.CFG(objects=[strstr_and_strncpy_amd64.loader.main_object], normalize=True)
    explorer.use_technique(angr.exploration_techniques.LoopSeer(cfg=cfg, bound=50))
    explorer.explore(find=[0x400657])
    s = explorer.found[0]
    result = s.solver.eval(s.memory.load(s.regs.rax, 15), cast_to=bytes)
    assert result == b"hi th hi there\x00"


def test_strstr():
    # auto_load_libs can't be disabled as the test cases failed.
    strstr_amd64 = angr.Project(
        os.path.join(test_location, "x86_64", "strstr"),
        load_options={"auto_load_libs": True},
        exclude_sim_procedures_list=["strstr"],
    )
    explorer = strstr_amd64.factory.simulation_manager()
    explorer.explore(find=[0x4005FB])
    s = explorer.found[0]
    result = s.solver.eval(s.memory.load(s.regs.rax, 9), cast_to=bytes)
    assert result == b"hi there\x00"


if __name__ == "__main__":
    test_strncpy()
    test_strncpy_size()
    test_strncpy_verify_null()
    test_strstr_and_strncpy()
    test_strstr()



# pylint:disable=missing-class-docstring,no-self-use,arguments-differ,unused-argument
import os

import angr

BIN_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "binaries")


def test_hook_symbol() -> None:
    """
    Test the hook_symbol (and related functions) useing the inet_ntoa simprocedure for functionality
    """
    bin_path = os.path.join(BIN_PATH, "tests", "x86_64", "inet_ntoa")
    proj = angr.Project(bin_path, auto_load_libs=False, use_sim_procedures=True)

    assert proj.is_symbol_hooked("inet_ntoa")
    assert not proj.is_symbol_hooked("not_expected_to_exist")

    original_hook = proj.symbol_hooked_by("inet_ntoa")

    assert isinstance(original_hook, angr.SIM_PROCEDURES["posix"]["inet_ntoa"])

    # No intention to call this, just checking hooking
    class FakeInetNtoa(angr.SimProcedure):
        def run(self, in_addr):
            return None

    fake_inet_ntoa = FakeInetNtoa()

    # test not allowing replacement
    proj.hook_symbol("inet_ntoa", fake_inet_ntoa, replace=False)
    assert proj.symbol_hooked_by("inet_ntoa") == original_hook

    # test allowing replacement
    proj.hook_symbol("inet_ntoa", fake_inet_ntoa, replace=True)
    assert proj.symbol_hooked_by("inet_ntoa") != original_hook
    assert proj.symbol_hooked_by("inet_ntoa") == fake_inet_ntoa



import sys

from angr import SimState, SimFile, SIM_PROCEDURES


def test_pwrite():
    pwrite = SIM_PROCEDURES["posix"]["pwrite64"]()

    state = SimState(arch="AMD64", mode="symbolic")
    simfile = SimFile("concrete_file", content="hello world!\n")
    state.fs.insert("test", simfile)
    fd = state.posix.open(b"test", 1)

    buf_addr = 0xD0000000
    state.memory.store(buf_addr, b"test!")
    pwrite.execute(state, arguments=[fd, buf_addr, 5, 6])

    simfd = state.posix.get_fd(fd)
    simfd.seek(0)
    res = 0xC0000000
    simfd.read(res, 13)
    data = state.solver.eval(state.mem[res].string.resolved, cast_to=bytes)

    assert data == b"hello test!!\n"

    state.posix.close(fd)


def test_pread():
    pwrite = SIM_PROCEDURES["posix"]["pread64"]()

    state = SimState(arch="AMD64", mode="symbolic")
    simfile = SimFile("concrete_file", content="hello world!\n")
    state.fs.insert("test", simfile)
    fd = state.posix.open(b"test", 1)

    buf1_addr = 0xD0000000
    buf2_addr = 0xD0001000
    pwrite.execute(state, arguments=[fd, buf1_addr, 6, 6])
    pwrite.execute(state, arguments=[fd, buf2_addr, 5, 0])

    data1 = state.solver.eval(state.mem[buf1_addr].string.resolved, cast_to=bytes)
    data2 = state.solver.eval(state.mem[buf2_addr].string.resolved, cast_to=bytes)

    assert data1 == b"world!"
    assert data2 == b"hello"

    state.posix.close(fd)


def main():
    g = globals()
    if len(sys.argv) > 1:
        f = "test_" + sys.argv[1]
        if f in g and hasattr(g[f], "__call__"):
            g[f]()
    else:
        for f, func in g.items():
            if f.startswith("test_") and hasattr(func, "__call__"):
                func()


if __name__ == "__main__":
    main()



from typing import Optional, Dict, Set, Iterable, Union, List
from functools import reduce

import networkx

import claripy
from cle.loader import Loader

from ...code_location import CodeLocation
from ...knowledge_plugins.key_definitions.atoms import Atom, MemoryLocation
from ...knowledge_plugins.key_definitions.definition import Definition
from ...knowledge_plugins.key_definitions.undefined import UNDEFINED
from ...knowledge_plugins.cfg import CFGModel
from .external_codeloc import ExternalCodeLocation


def _is_definition(node):
    return isinstance(node, Definition)


class DepGraph:
    """
    The representation of a dependency graph: a directed graph, where nodes are definitions, and edges represent uses.

    Mostly a wrapper around a <networkx.DiGraph>.
    """

    def __init__(self, graph: Optional[networkx.DiGraph] = None):
        """
        :param graph: A graph where nodes are definitions, and edges represent uses.
        """
        # Used for memoization of the `transitive_closure` method.
        self._transitive_closures: Dict = {}

        if graph and not all(map(_is_definition, graph.nodes)):
            raise TypeError("In a DepGraph, nodes need to be <%s>s." % Definition.__name__)

        self._graph: networkx.DiGraph = graph if graph is not None else networkx.DiGraph()

    @property
    def graph(self) -> networkx.DiGraph:
        return self._graph

    def add_node(self, node: Definition) -> None:
        """
        :param node: The definition to add to the definition-use graph.
        """
        self._graph.add_node(node)

    def add_edge(self, source: Definition, destination: Definition, **labels) -> None:
        """
        The edge to add to the definition-use graph. Will create nodes that are not yet present.

        :param source: The "source" definition, used by the "destination".
        :param destination: The "destination" definition, using the variable defined by "source".
        :param labels: Optional keyword arguments to represent edge labels.
        """
        self._graph.add_edge(source, destination, **labels)

    def nodes(self) -> networkx.classes.reportviews.NodeView:
        return self._graph.nodes()

    def predecessors(self, node: Definition) -> networkx.classes.reportviews.NodeView:
        """
        :param node: The definition to get the predecessors of.
        """
        return self._graph.predecessors(node)

    def transitive_closure(self, definition: Definition) -> networkx.DiGraph:
        """
        Compute the "transitive closure" of a given definition.
        Obtained by transitively aggregating the ancestors of this definition in the graph.

        Note: Each definition is memoized to avoid any kind of recomputation across the lifetime of this object.

        :param definition:  The Definition to get transitive closure for.
        :return:            A graph of the transitive closure of the given definition.
        """

        def _transitive_closure(
            def_: Definition,
            graph: networkx.DiGraph,
            result: networkx.DiGraph,
            visited: Optional[Set[Definition]] = None,
        ):
            """
            Returns a joint graph that comprises the transitive closure of all defs that `def_` depends on and the
            current graph `result`. `result` is updated.
            """
            if def_ in self._transitive_closures:
                closure = self._transitive_closures[def_]
                # merge closure into result
                result.add_edges_from(closure.edges())
                return result

            if def_ not in graph:
                return result

            predecessors = list(graph.predecessors(def_))

            result.add_node(def_)
            for pred in predecessors:
                edge_data = graph.get_edge_data(pred, def_)
                if edge_data is None:
                    result.add_edge(pred, def_)
                else:
                    result.add_edge(pred, def_, **edge_data)

            visited = visited or set()
            visited.add(def_)
            predecessors_to_visit = set(predecessors) - set(visited)

            closure = reduce(
                lambda acc, def0: _transitive_closure(def0, graph, acc, visited), predecessors_to_visit, result
            )

            self._transitive_closures[def_] = closure
            return closure

        return _transitive_closure(definition, self._graph, networkx.DiGraph())

    def contains_atom(self, atom: Atom) -> bool:
        return any(map(lambda definition: definition.atom == atom, self.nodes()))

    def add_dependencies_for_concrete_pointers_of(
        self, values: Iterable[Union[claripy.ast.Base, int]], definition: Definition, cfg: CFGModel, loader: Loader
    ):
        """
        When a given definition holds concrete pointers, make sure the <MemoryLocation>s they point to are present in
        the dependency graph; Adds them if necessary.

        :param values:
        :param definition: The definition which has data that can contain concrete pointers.
        :param cfg: The CFG, containing information about memory data.
        :param loader:
        """
        assert definition in self.nodes(), "The given Definition must be present in the given graph."

        known_predecessor_addresses: List[Union[int, claripy.ast.Base]] = list(
            # Needs https://github.com/python/mypy/issues/6847
            map(
                lambda definition: definition.atom.addr,  # type: ignore
                filter(lambda p: isinstance(p.atom, MemoryLocation), self.predecessors(definition)),
            )
        )

        # concretize addresses where possible
        concrete_known_pred_addresses = []
        for address in known_predecessor_addresses:
            if isinstance(address, claripy.ast.Base):
                if address.concrete:
                    concrete_known_pred_addresses.append(address._model_concrete.value)
            else:
                concrete_known_pred_addresses.append(address)

        unknown_concrete_addresses: Set[int] = set()
        for v in values:
            if isinstance(v, claripy.ast.Base) and v.concrete:
                v = v._model_concrete.value
            if isinstance(v, int):
                if v not in concrete_known_pred_addresses:
                    unknown_concrete_addresses.add(v)

        for address in unknown_concrete_addresses:
            data_at_address = cfg.memory_data.get(address, None) if cfg is not None else None

            if data_at_address is None or data_at_address.sort not in ["string", "unknown"]:
                continue

            section = loader.main_object.find_section_containing(address)
            read_only = False if section is None else not section.is_writable
            code_location = CodeLocation(0, 0, info={"readonly": True}) if read_only else ExternalCodeLocation()

            def _string_and_length_from(data_at_address):
                if data_at_address.content is None:
                    return UNDEFINED, data_at_address.size
                else:
                    return data_at_address.content.decode("utf-8"), data_at_address.size + 1

            _, string_length = _string_and_length_from(data_at_address)

            memory_location_definition = Definition(
                MemoryLocation(address, string_length),
                code_location,
            )

            self.graph.add_edge(memory_location_definition, definition)
